<?php

require_once "class.ilFormATestSessionSet.php";

/**
 * Class ilFormATestSession
 * Date: 16.12.13
 * Time: 09:59
 * @author Thomas Joußen <tjoussen@databay.de>
 */
class ilFormATestSession
{

	const STATUS_FINISHED     = "finished";
	const STATUS_INTERRUPTED  = "interrupted";
	const STATUS_NOT_FINISHED = "not_finished";
	const STATUS_NOT_STARTED  = "not_started";
	const STATUS_IN_PROGRESS  = "in_progress";

	/**
	 * @var int
	 */
	protected $obj_id;

	/**
	 * @var int
	 */
	protected $ref_id;

	/**
	 * @var null|int
	 */
	protected $session_id = null;

	/**
	 * @var null|int
	 */
	protected $active_id = null;

	/**
	 * @var int
	 */
	protected $user_id;

	/**
	 * @var int
	 */
	protected $test_id;

	/**
	 * @var int
	 */
	protected $pass;

	/**
	 * @var DateTime
	 */
	protected $start_date;

	/**
	 * @var DateTime
	 */
	protected $end_date;

	/**
	 * @var string
	 */
	protected $status;

	/**
	 * @var bool
	 */
	protected $fixate_solution;

	/**
	 * @var ilDB
	 */
	protected $db;

	/**
	 * @var ilFormATestSessionSet[]
	 */
	protected $sets;

	/**
	 * @var ilFormATestSessionSet
	 */
	protected $current_position;

	/**
	 * @param ilObjFormATest $test
	 */
	public function __construct($test)
	{
		global $ilDB, $ilUser;

		$this->db      = $ilDB;
		$this->user_id = $ilUser->getId();
		$this->test_id = $test->getTestId();
		$this->obj_id = $test->getId();
		$this->ref_id = $test->getRefId();
		$this->fixate_solution = false;
	}

	/**
	 * @return bool
	 */
	public function activeIdExists()
	{
		return $this->active_id != null;
	}

	/**
	 * @return bool
	 */
	public function isPassFinished()
	{
		return $this->end_date != null;
	}

	/**
	 * @return bool
	 */
	public function isSetFinished()
	{
		return $this->current_position->isFinished();
	}

	/**
	 * @return bool
	 */
	public function isFirstStep()
	{
		return $this->current_position->getStep() <= 1;
	}

	/**
	 * @return mixed
	 */
	public function hasNextSet()
	{
		return array_search($this->current_position, $this->sets) < count($this->sets) - 1;
	}

	/**
	 * @return bool
	 */
	public function isJump()
	{
		return $this->current_position->isJump();
	}

	public function unsetJump()
	{
		$this->current_position->setJump(false);
	}


	public function markSolutionToFixate()
	{
		$this->fixate_solution = true;
	}

	/**
	 * @return bool
	 */
	public function allSetsFinished()
	{
		$status = true;
		foreach($this->sets as $set)
		{
			$status = $status && $set->isFinished();
		}

		return $status;
	}

	/**
	 * @return string
	 */
	public function getToken()
	{
		return md5("xtsf_" . $this->test_id . $this->session_id/* . $this->getNodeId()*/);
	}

	/**
	 * Create a new test session for a user in a test
	 */
	public function create()
	{
		$this->active_id = $this->db->nextId("tst_active");

		$this->db->insert(
			"tst_active",
			array(
				"active_id" => array("integer", $this->active_id),
				"user_fi"   => array("integer", $this->user_id),
				"test_fi"   => array("integer", $this->test_id),
			)
		);

		$this->createNewPass();
	}

	/**
	 * Create a new session pass for a user in a test
	 */
	public function createNewPass()
	{
		$this->initNewPass();

		$this->db->update(
			"tst_active",
			array(
				"tries"  => array("integer", $this->pass),
				"tstamp" => array("integer", $this->start_date->getTimestamp())
			),
			array("active_id" => array("integer", $this->active_id))
		);

		$this->db->insert(
			"rep_robj_xtsf_session",
			array(
				"session_id"   => array("integer", $this->session_id),
				"active_id"    => array("integer", $this->active_id),
				"start_date"   => array("integer", $this->start_date->getTimestamp()),
				"current_pass" => array("integer", $this->pass),
				"status"       => array("text", self::STATUS_NOT_FINISHED)
			)
		);
	}

	/**
	 * @param array $sets
	 */
	public function read($sets)
	{
		$result = $this->db->queryF(
			"SELECT sess.* FROM tst_active tsta
			INNER JOIN rep_robj_xtsf_session sess ON sess.active_id = tsta.active_id AND sess.current_pass = tsta.tries
			WHERE tsta.user_fi = %s AND tsta.test_fi = %s",
			array("integer", "integer"),
			array($this->user_id, $this->test_id)
		);

		if($result->numRows())
		{
			$this->bindData($result);
		}
		$this->initSessionSet($sets);
	}

	/**
	 * @return int|null
	 */
	public function getActiveId()
	{
		return $this->active_id;
	}

	/**
	 * @return int|null
	 */
	public function getSessionId()
	{
		return $this->session_id;
	}

	/**
	 * @return int
	 */
	public function getUserId()
	{
		return $this->user_id;
	}

	/**
	 * Updates the current session
	 */
	public function update()
	{
		$date = null;
		if($this->status == self::STATUS_FINISHED)
		{
			$this->end_date = new DateTime("now");
			$this->updateLP();
			$date = $this->end_date->getTimestamp();
		}

		$this->db->update(
			"rep_robj_xtsf_session",
			array(
				"end_date" => array("integer", $date),
				"status"   => array("text", $this->status)
			),
			array(
				"session_id" => array("integer", $this->session_id)
			)
		);
	}

	/**
	 * Does a new session step
	 *
	 * @param int $node_id
	 */
	public function step($node_id)
	{
		$this->current_position->setNodeId($node_id);
		$this->current_position->save();
	}

	/**
	 * Initialize the current position for the session
	 */
	public function initCurrent()
	{
		foreach($this->sets as $set)
		{
			if($this->isCurrentSet($set) || ($set->isNotStarted() && $this->current_position->isFinished()))
			{
				$this->current_position = $set;
			}
		}

		if($this->current_position->getNodeId() == null)
		{
			$this->current_position->init();
		}
	}

	/**
	 * @return int
	 */
	public function getNodeId()
	{
		return $this->current_position->getNodeId();
	}

	/**
	 * @return int
	 */
	public function getStep()
	{
		return $this->current_position->getStep();
	}

	/**
	 * @return int
	 */
	public function getPass()
	{
		return $this->pass;
	}

	/**
	 * @param string $status
	 */
	public function setStatus($status)
	{
		$this->status = $status;
	}

	/**
	 * @param int      $node_id
	 * @param string   $condition
	 * @param null|int $condition_id
	 */
	public function updateCondition($node_id, $condition, $condition_id = null)
	{
		$this->current_position->setConditionNodeId($node_id);
		$this->current_position->setCondition($condition);
		$this->current_position->setConditionId($condition_id);
		$this->current_position->update();
	}

	/**
	 * @param string $status
	 */
	public function updateSetStatus($status)
	{
		$this->current_position->setStatus($status);
		$this->current_position->update();
		if($this->current_position->isFinished())
		{
			$this->current_position->setNodeId(NULL);
		}
	}

	/**
	 * @param int $index
	 */
	public function removeStepsUntil($index)
	{
		$this->current_position->removeUntil($index);
		$this->current_position->read();
	}

	/**
	 * @param int $step
	 */
	public function removeFeedbackByStep($step)
	{
		$this->db->manipulateF('
		DELETE rep_robj_xtsf_fbt FROM rep_robj_xtsf_fbt 
			WHERE session_id = %s 
			AND set_id = %s
		AND step = %s ',
			array('integer', 'integer', 'integer'),
			array($this->session_id, $this->current_position->getSetId(), $step));
	}
	
	/**
	 * @param ilQuestionSetPoolNode $node
	 * @param ilQuestionSetPoolNode $current_node
	 */
	public function changePath($node, $current_node)
	{
		if($node->getPath()->getSet()->getId() == $current_node->getPath()->getSet()->getId())
		{
			$newNode = $current_node->getPath()->getNodes()->getNodeByQuestionId($node->getQuestion()->getId());
			$newPreviousNode = $current_node->getPath()->getNodes()->getPreviousNode($newNode);
			$previousNode = $node->getPath()->getNodes()->getPreviousNode($node);

			if($newPreviousNode != null && $previousNode != null)
			{
				$this->db->update(
					"rep_robj_xtsf_tracking",
					array(
						"matched_condition_node_id" => array("integer", $newPreviousNode->getId())
					),
					array(
						"session_fi" => array("integer", $this->session_id),
						"set_fi" => array("integer", $this->getSetId()),
						"node_fi" => array("integer", $previousNode->getId())
					)
				);
			}

			//Update node
			$this->db->update(
				"rep_robj_xtsf_tracking",
				array(
					"node_fi" => array("integer", $newNode->getId())
				),
				array(
					"session_fi" => array("integer", $this->session_id),
					"step" => array("integer", $this->getStep()),
					"set_fi" => array("integer", $this->getSetId()),
				)
			);
		}
	}


	/**
	 * Swap the current position to the delivered set
	 *
	 * @param int $set_id
	 */
	public function swapCurrent($set_id)
	{
		if($this->current_position != null && $this->current_position->getSetId() != $set_id && !$this->current_position->isFinished())
		{
			$this->current_position->setStatus(self::STATUS_INTERRUPTED);
			$this->current_position->update();
		}

		foreach($this->sets as $set)
		{
			if($set->getSetId() == $set_id)
			{
				$this->current_position = $set;
				if($this->current_position->getNodeId() == null)
				{
					$this->current_position->init();
					$this->step($this->getNodeId());
				}
				else
				{
					$this->updateSetStatus(self::STATUS_IN_PROGRESS);
				}
			}
		}

	}

	/**
	 * @param null|int $set_id
	 *
	 * @return string
	 */
	public function getSetStatus($set_id = null)
	{
		if($set_id == null)
		{
			return $this->current_position->getStatus();
		}
		else
		{
			foreach($this->sets as $set)
			{
				if($set->getSetId() == $set_id)
				{
					return $set->getStatus();
				}
			}
		}

		return self::STATUS_NOT_STARTED;
	}

	/**
	 * @return int
	 */
	public function getSetId()
	{
		return $this->current_position->getSetId();
	}

	/**
	 * @param ilQuestionSetPoolNode $node
	 *
	 * @return array
	 */
	public function getStepsToDelete($node)
	{
		$result = $this->db->queryF(
			"SELECT DISTINCT track.step, node.*, qst.title, qs_qst.question_index FROM rep_robj_xqsp_node node
			INNER JOIN rep_robj_xtsf_tracking track ON track.node_fi = node.id
			INNER JOIN qpl_questions qst ON qst.question_id = node.question_fi
			INNER JOIN rep_robj_xtsf_session sess ON sess.session_id = track.session_fi
			INNER JOIN tst_solutions sol ON sol.question_fi = qst.question_id AND sol.step = track.step AND sol.active_fi = sess.active_id AND sol.pass = sess.current_pass
			INNER JOIN rep_robj_xqsp_qs_qst qs_qst ON qs_qst.question_fi = qst.question_id
			WHERE track.step > (
				SELECT track2.step FROM rep_robj_xtsf_tracking track2
				WHERE track2.node_fi = %s AND track2.session_fi = %s
			) AND track.set_fi = %s AND track.session_fi = %s
			ORDER BY track.step",
			array("integer", "integer", "integer", "integer"),
			array($node->getId(), $this->getSessionId(), $node->getPath()->getSet()->getId(), $this->getSessionId())
		);

		$data = array();
		while($row = $this->db->fetchAssoc($result))
		{
			$data[] = $row;
		}

		return $data;
	}

	/**
	 * @param ilQuestionSetPoolNode $node
	 *
	 * @return bool
	 */
	public function stepHasSolution($node)
	{
		$result = $node->getQuestion()->getUserQuestionResult($this->active_id, $this->pass);

		return count($result->getSolutions()) > 0;
	}

	/**
	 * @param ilQuestionSetPoolNode $node
	 *
	 * @return bool
	 */
	public function hasStepsToDelete($node)
	{
		return count($this->getStepsToDelete($node)) > 0;
	}

	/**
	 * Increments the pass counter
	 */
	protected function setPassCounter()
	{
		if($this->pass == null)
		{
			$this->pass = 0;
		}
		else
		{
			$this->pass++;
		}
	}

	/**
	 * @param ilFormATestSessionSet $session_set
	 *
	 * @return bool
	 */
	protected function isCurrentSet($session_set)
	{
		return
			$this->current_position == null ||
			(($session_set->isCurrent() || $session_set->isFinished()) && !$this->current_position->isCurrent()) /*||
			($session_set->isNotStarted() && $this->current_position->isFinished())*/
			;
	}

	/**
	 * Initialize a new Pass. Sets default data
	 */
	protected function initNewPass()
	{
		$this->session_id       = $this->db->nextId("rep_robj_xtsf_session");
		$this->start_date       = new DateTime("now");
		$this->end_date         = null;
		$this->current_position = null;
		$this->setPassCounter();

		foreach($this->sets as $key => $set)
		{
			$session_set = new ilFormATestSessionSet($set->getSetId());
			$session_set->setSessionId($this->session_id);
			$session_set->setStatus(self::STATUS_NOT_STARTED);
			$session_set->create();
			$this->sets[$key] = $session_set;
		}
	}

	/**
	 * @param array $sets
	 */
	protected function initSessionSet($sets)
	{
		foreach($sets as $set)
		{
			$session_set = new ilFormATestSessionSet($set["id"]);
			$session_set->setSessionId($this->session_id);
			$session_set->read();
			$this->sets[] = $session_set;

			if($this->isCurrentSet($session_set))
			{
				$this->current_position = $session_set;
			}
		}
	}

	/**
	 * @param object $result
	 */
	protected function bindData($result)
	{
		$row              = $this->db->fetchAssoc($result);
		$this->session_id = $row["session_id"];
		$this->active_id  = $row["active_id"];
		$this->pass       = $row["current_pass"];
		$this->start_date = $row["start_date"];
		$this->end_date   = $row["end_date"];
		$this->status     = $row["status"];
	}


	/**
	 * @return bool
	 */
	public function reachedDeadlock()
	{
		if($this->isJump())
		{
			return $this->current_position->isDeadlock();
		}

		return false;
	}

	/**
	 * @return bool
	 */
	public function isDeadlockPossible()
	{
		return $this->current_position->isDeadlockPossible();
	}

	/**
	 * Saved the history of the test to track all steps and results
	 */
	public function saveHistory()
	{
		$this->db->queryF(
			"INSERT INTO rep_robj_xtsf_trck_his (session_fi, step, node_fi, set_fi, matched_condition_node_id, matched_condition, matched_condition_id)
			SELECT trck.session_fi, (
				SELECT (CASE WHEN MAX(trck2.step) IS NULL THEN 1 ELSE MAX(trck2.step)+1 END) step FROM rep_robj_xtsf_trck_his trck2 WHERE trck2.session_fi = %s AND trck2.set_fi = %s
			), trck.node_fi, trck.set_fi, trck.matched_condition_node_id, trck.matched_condition, trck.matched_condition_id
			FROM rep_robj_xtsf_tracking trck
			WHERE trck.session_fi = %s AND trck.step = %s AND trck.set_fi = %s",
			array("integer","integer","integer","integer","integer"),
			array($this->getSessionId(),$this->getSetId(),$this->getSessionId(),$this->getStep(),$this->getSetId())
		);

		if($this->fixate_solution)
		{
			$this->db->queryF(
				"UPDATE rep_robj_xtsf_solut solut
				INNER JOIN rep_robj_xqsp_node node ON node.question_fi = solut.question_fi
				SET fixated = %s
				WHERE active_fi = %s AND pass = %s AND node.id = %s",
				array("integer","integer", "integer", "integer"),
				array(0, $this->getActiveId(), $this->getPass(), $this->getNodeId())
			);
		}


		$this->db->queryF(
			"INSERT INTO rep_robj_xtsf_solut (solution_id, active_fi, question_fi, points, pass, tstamp, value1, value2, step, fixated)
			SELECT tst_solutions.solution_id, tst_solutions.active_fi, tst_solutions.question_fi, tst_solutions.points, tst_solutions.pass, tst_solutions.tstamp, tst_solutions.value1, tst_solutions.value2,
			(
				SELECT MAX(trck2.step) FROM rep_robj_xtsf_trck_his trck2 WHERE trck2.set_fi = %s AND trck2.node_fi = %s AND trck2.session_fi = %s
			), %s
			FROM tst_solutions
			INNER JOIN rep_robj_xqsp_node node ON node.question_fi = tst_solutions.question_fi
			WHERE tst_solutions.active_fi = %s AND tst_solutions.pass = %s AND tst_solutions.step = %s AND node.id = %s",
			array("integer","integer","integer","integer","integer","integer","integer","integer"),
			array(
				$this->getSetId(),
				$this->getNodeId(),
				$this->getSessionId(),
				$this->fixate_solution,
				$this->getActiveId(),
				$this->getPass(),
				$this->getStep(),
				$this->getNodeId()
			)
		);
	}

	/**
	 * @param null|int $node_id
	 *
	 * @return bool
	 */
	public function hasFixatedSolution($node_id = null)
	{
		return $this->current_position->hasFixatedSolution($node_id);
	}

	public function handleDeadlock()
	{
		$this->db->queryF(
			"INSERT INTO tst_solutions (solution_id, active_fi, question_fi, points, pass, tstamp, value1, value2, step)
			SELECT rep_robj_xtsf_solut.solution_id, rep_robj_xtsf_solut.active_fi, rep_robj_xtsf_solut.question_fi,
			 rep_robj_xtsf_solut.points, rep_robj_xtsf_solut.pass, rep_robj_xtsf_solut.tstamp,
			  rep_robj_xtsf_solut.value1, rep_robj_xtsf_solut.value2,%s
			FROM rep_robj_xtsf_solut
			INNER JOIN rep_robj_xqsp_node node ON node.question_fi = rep_robj_xtsf_solut.question_fi
			WHERE rep_robj_xtsf_solut.active_fi = %s AND rep_robj_xtsf_solut.pass = %s AND rep_robj_xtsf_solut.fixated = %s AND node.id = %s",
			array("integer","integer","integer","integer","integer"),
			array(
				$this->getStep(),
				$this->getActiveId(),
				$this->getPass(),
				1,
				$this->getNodeId()
			)
		);
	}

	/**
	 * @param int $node_id
	 */
	public function updateTrackNodeId($node_id)
	{
		$this->current_position->updateTrackNodeId($node_id);
	}

	/**
	 * @param int $active_id
	 *
	 * @return int
	 */
	public static function getUserIdByActiveId($active_id)
	{
		/** @var ilDB $ilDB */
		global $ilDB;

		$result = $ilDB->queryF(
			"SELECT user_fi FROM tst_active WHERE active_id = %s",
			array("integer"),
			array($active_id)
		);
		$row    = $ilDB->fetchAssoc($result);

		return $row["user_fi"];
	}

	/**
	 * @param int $session_id
	 *
	 * @return int
	 */
	public static function getUserIdBySessionId($session_id)
	{
		/** @var ilDB $ilDB */
		global $ilDB;

		$ilDB->setLimit(1, 0);
		$result = $ilDB->queryF(
			"SELECT user_fi FROM tst_active
			INNER JOIN rep_robj_xtsf_session ON rep_robj_xtsf_session.active_id = tst_active.active_id
			WHERE rep_robj_xtsf_session.session_id = %s",
			array("integer"),
			array($session_id)
		);
		$row    = $ilDB->fetchAssoc($result);

		return $row["user_fi"];
	}

	/**
	 * @param int $session_id
	 *
	 * @return int
	 */
	public static function getActiveIdBySessionId($session_id)
	{
		/** @var ilDB $ilDB */
		global $ilDB;

		$ilDB->setLimit(1, 0);
		$result = $ilDB->queryF(
			"SELECT active_id FROM rep_robj_xtsf_session WHERE session_id = %s",
			array("integer"),
			array($session_id)
		);
		$row    = $ilDB->fetchAssoc($result);

		return $row["active_id"];
	}

	/**
	 * @param $session_id
	 *
	 * @return mixed
	 */
	public static function getPassBySessionId($session_id)
	{
		/** @var ilDB $ilDB */
		global $ilDB;

		$ilDB->setLimit(1, 0);
		$result = $ilDB->queryF(
			"SELECT current_pass FROM rep_robj_xtsf_session WHERE session_id = %s",
			array("integer"),
			array($session_id)
		);
		$row    = $ilDB->fetchAssoc($result);

		return $row["active_id"];
	}

	/**
	 * @param int $active_id
	 * @param int $pass
	 *
	 * @return int
	 */
	public static function getSessionIdByActiveIdAndPass($active_id, $pass)
	{
		/** @var ilDB $ilDB */
		global $ilDB;

		$result = $ilDB->queryF(
			"SELECT session_id FROM rep_robj_xtsf_session WHERE active_id = %s and current_pass = %s",
			array("integer", "integer"),
			array($active_id, $pass)
		);

		$row = $ilDB->fetchAssoc($result);

		return $row["session_id"];
	}

	public function updateLP()
	{
		require_once "./Modules/Test/classes/class.ilObjTestAccess.php";
		require_once "./Services/Tracking/classes/status/class.ilLPStatusEvent.php";
		require_once "./Services/Tracking/classes/class.ilLPStatusWrapper.php";
		require_once "./Services/Tracking/classes/class.ilLearningProgress.php";

		ilLPStatusWrapper::_updateStatus(
			$this->obj_id,
			$this->user_id
		);
	}

	public function trackProgress()
	{
		require_once "./Modules/Test/classes/class.ilObjTestAccess.php";
		require_once "./Services/Tracking/classes/class.ilChangeEvent.php";
		require_once "./Services/Tracking/classes/status/class.ilLPStatusEvent.php";
		require_once "./Services/Tracking/classes/class.ilLPStatusWrapper.php";
		require_once "./Services/Tracking/classes/class.ilLearningProgress.php";

		ilLearningProgress::_tracProgress(
			$this->user_id,
			$this->obj_id,
			$this->ref_id,
			'xtsf'
		);
	}



}